/**
 *    Free Tide Mini is a free harmonic tide predictor
 *    Copyright (C) 2008  Regis COSNIER
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package freetide.mini;

import freetide.data.HarmonicFile;
import freetide.data.Location;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import javax.microedition.lcdui.*;

/**
 * @author Dgis
 */
public class TideCanvas extends Canvas implements CommandListener {
	
	private Location location = null;
	private Location lastLocation = null;
	private int numTideEvent = 6;
	private long timeOffset = 0;
	private boolean forceRepaint = false;
	private Image frame = null;
	private String waitingMessage = LocalizationSupport.getMessage("TEXT_WAITING");
	private Font font = null;
	private int fontHeight = 0;
	private int unit = HarmonicFile.UNIT_METER;
	private int language = Configuration.LANGUAGE_ENGLISH;
	private int clockType = Configuration.CLOCK_24H;
	
    /**
     * constructor
     */
    public TideCanvas()
	{
		font = Font.getFont(Font.FACE_MONOSPACE, Font.STYLE_PLAIN, Font.SIZE_SMALL);
		fontHeight = font.getBaselinePosition();
    } 
    
	public void setLocation(Location location)
	{
		this.location = location;
		timeOffset = 0;
		repaint();
	}
	
	
	public void setWaitingMessage(String message)
	{
		waitingMessage = message;
	}
	
	public void setUnit(int unit) { this.unit = unit; forceRepaint = true; }
	public void setLanguage(int language) { this.language = language; forceRepaint = true; }
	public void setClockType(int clockType) { this.clockType = clockType; forceRepaint = true; }
	
	public static String formatGMTDateLong(Date date, int clockType)
	{
		Calendar calendar = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
		calendar.setTime(date);
		return formatCalendarLong(calendar, clockType);
	}
	
	public static String formatCalendarLong(Calendar calendar, int clockType)
	{
		return calendar.get(Calendar.YEAR) + "/" + formatInteger(calendar.get(Calendar.MONTH) + 1, 2) +
				"/" + formatInteger(calendar.get(Calendar.DATE), 2) +
				" " + (clockType == Configuration.CLOCK_24H ?
					formatInteger(calendar.get(Calendar.HOUR_OF_DAY), 2) :
					formatInteger(calendar.get(Calendar.HOUR), 2)) +
				":" + formatInteger(calendar.get(Calendar.MINUTE), 2) +
				(clockType == Configuration.CLOCK_24H ? "" : (calendar.get(Calendar.AM_PM) == Calendar.AM ? " AM" : " PM"));
	}
	
	public static String formatCalendarShort(Calendar calendar, int clockType)
	{
		return (clockType == Configuration.CLOCK_24H ?
					formatInteger(calendar.get(Calendar.HOUR_OF_DAY), 2) :
					"" + calendar.get(Calendar.HOUR)) +
				":" + formatInteger(calendar.get(Calendar.MINUTE), 2) +
				(clockType == Configuration.CLOCK_24H ? "" : (calendar.get(Calendar.AM_PM) == Calendar.AM ? "AM" : "PM"));
	}
	
	public static String formatFloat(float value, int decimalPlaces)
	{
		String doubleString = "" + value;
		return doubleString.substring(0, (doubleString.indexOf(".") + decimalPlaces + 1));
	}
	
	public static String formatTideType(int tideType)
	{
		if(tideType == Location.NextTide.LOW_TIDE)
			return LocalizationSupport.getMessage("TEXT_TIDE_LOW");
		else if(tideType == Location.NextTide.HIGH_TIDE)
			return LocalizationSupport.getMessage("TEXT_TIDE_HIGH");
		return "";
	}

	public static String formatInteger(int value, int decimalPlaces)
	{
		String result  = "" + value;
		if(result.length() < decimalPlaces)
		{
			int missingZero = decimalPlaces - result.length();
			for (int i = 0; i < missingZero; i++)
				result = "0" + result;
		}
		return result;
	}
	
    /**
     * paint
     */
    public void paint(Graphics gf) {
		
		if(location != null && location.isLoaded())
		{
			if(lastLocation != location || forceRepaint)
			{
				forceRepaint = false;
				
				frame = Image.createImage(getWidth(), getHeight());
				Graphics g = frame.getGraphics();    
				
				g.setColor(0x000000); //g.setColor(0xC9ECFF); //g.setColor(0x708CFF);
				g.fillRect(0, 0, getWidth(), getHeight());
				g.setColor(0xFFFFFF);

				g.drawString(location.toString(), 0, 0, Graphics.TOP | Graphics.LEFT);
		
				g.setFont(font);
				
				String unitString = (unit == HarmonicFile.UNIT_METER ?
					LocalizationSupport.getMessage("TEXT_UNIT_METER") :
					LocalizationSupport.getMessage("TEXT_UNIT_FEET"));

				TimeZone displayTimeZone = TimeZone.getDefault(); //getTimeZone("GMT+1");
				Calendar currentlocalCalendar = Calendar.getInstance(displayTimeZone);
				Date currentGMTDate = currentlocalCalendar.getTime();
				Date currentLocalDate = new Date(currentGMTDate.getTime() - displayTimeZone.getRawOffset());
				long nowGMTTime = currentLocalDate.getTime() / 1000;
				float currentHeight = location.time2asecondary(nowGMTTime);
				currentHeight = HarmonicFile.converterUnit(currentHeight, HarmonicFile.UNIT_METER, unit);
				//float currentNormalizedHeight = location.time2secondary(nowGMTTime);
				g.drawString(LocalizationSupport.getMessage("TEXT_NOW") + formatGMTDateLong(currentGMTDate, clockType) + " " + formatFloat(currentHeight, 1) + " " + unitString,
						0, (int)(1.8f * fontHeight), Graphics.TOP | Graphics.LEFT);
				
					
				// Display the tide calendar
			
				class TideSummary
				{
					String shortLocalDate;
					long gmtTime;
					int tideType;
					float tideHeight;
					//float tideNormalizedHeight;
				}
				
				TideSummary[] tideEvents = new TideSummary[numTideEvent];
				long startGMTTime = nowGMTTime - 7 * 3600 + timeOffset; // Display start: now - 7 hours
				Location.NextTide nextTide = location.new NextTide(startGMTTime);
				for (int i = 0; i < numTideEvent; i++)
				{
					location.nextTide(nextTide);
					Date tideGMTDate = nextTide.getTideGMTDate();
					currentlocalCalendar.setTime(tideGMTDate);
					tideEvents[i] = new TideSummary();
					tideEvents[i].shortLocalDate = formatCalendarShort(currentlocalCalendar, clockType);
					tideEvents[i].gmtTime = tideGMTDate.getTime() / 1000;
					tideEvents[i].tideHeight = nextTide.getTideHeight();
					tideEvents[i].tideHeight = HarmonicFile.converterUnit(tideEvents[i].tideHeight, HarmonicFile.UNIT_METER, unit);
					//tideEvents[i].tideNormalizedHeight = nextTide.getTideNormalizedHeight();
					tideEvents[i].tideType = nextTide.getTideType();
					g.drawString(formatCalendarLong(currentlocalCalendar, clockType) + "  " + formatTideType(tideEvents[i].tideType) + " " + formatFloat(tideEvents[i].tideHeight, 1) + " " + unitString,
							0, (i + 3) * fontHeight, Graphics.TOP | Graphics.LEFT);
				}

				// Display the tide graph
				g.setColor(0x0000ff);

				long endGMTTime = nextTide.getTideGMTTime() + 3 * 3600;

				int yGraph = (numTideEvent + 4) * fontHeight; //this.getHeight() * 2 / 3;
				int graphWidth = this.getWidth();
				int graphHeight = this.getHeight() - yGraph; //this.getHeight() / 3;
				int graphHeight2 = graphHeight / 2;

				int numStep = 60;
				float scaleX = (float)graphWidth / numStep;
				float scaleY = (float)graphHeight / 2.2f; // 2.2 = (1 - -1) + 0.2
				long timeStep = (endGMTTime - startGMTTime) / numStep;
				long graphGMTTime = startGMTTime;
				float currentTideHeight = 0.0f, lastTideHeight = location.time2secondary (graphGMTTime);
				for (int i = 1; i < numStep; i++)
				{
					graphGMTTime += timeStep;
					currentTideHeight = location.time2secondary (graphGMTTime);

					g.drawLine(	(int)(scaleX * (i - 1)),	yGraph + graphHeight2 - (int)(scaleY * lastTideHeight),
								(int)(scaleX * i),			yGraph + graphHeight2 - (int)(scaleY * currentTideHeight));

					lastTideHeight = currentTideHeight;
				}

				// Display the lowest tide line
				g.setColor(0x00ff00);
				int yLowestTide = yGraph + graphHeight2 - (int)(scaleY * -1);
				g.drawLine(	0, yLowestTide, getWidth() - 1, yLowestTide);

				// Display the highest tide line
				g.setColor(0xff0000);
				int yHighestTide = yGraph + graphHeight2 - (int)(scaleY * 1);
				g.drawLine(	0, yHighestTide, getWidth() - 1, yHighestTide);

				// Display the current time mark
				g.setColor(0xC300FF);
				int xTime = (int)(scaleX * (nowGMTTime - startGMTTime) / (float)timeStep);
				g.drawLine(xTime, yLowestTide - 10, xTime, yHighestTide + 10);
				
				// Display the tide events on the graph
				g.setColor(0xFFFFFF);
				for (int i = 0; i < numTideEvent; i++)
				{
					boolean isLow = tideEvents[i].tideType == Location.NextTide.LOW_TIDE;
					int xTide = (int)(scaleX * (tideEvents[i].gmtTime - startGMTTime) / (float)timeStep);
					if(isLow)
					{
						int yTide = yLowestTide - 1;
						int yTide2 = yTide - fontHeight;
						g.drawString(tideEvents[i].shortLocalDate, xTide, yTide, Graphics.BOTTOM | Graphics.HCENTER);
						g.drawString(formatFloat(tideEvents[i].tideHeight, 1) + unitString, xTide, yTide2, Graphics.BOTTOM | Graphics.HCENTER);
					}
					else
					{
						int yTide = yHighestTide + 1;
						int yTide2 = yTide + fontHeight;
						g.drawString(tideEvents[i].shortLocalDate, xTide, yTide, Graphics.TOP | Graphics.HCENTER);
						g.drawString(formatFloat(tideEvents[i].tideHeight, 1) + unitString, xTide, yTide2, Graphics.TOP | Graphics.HCENTER);
					}
				}
				
				
				lastLocation = location;
			}
			gf.drawImage(frame, 0, 0, Graphics.TOP | Graphics.LEFT);
		}
		else
		{
			gf.setColor(0x000000);
			gf.fillRect(0, 0, getWidth(), getHeight());
			gf.setColor(0xffffff);
	        gf.drawString(waitingMessage, 0, 0, Graphics.TOP | Graphics.LEFT);
		}
    }
    
    /**
     * Called when a key is pressed.
     */
    protected  void keyPressed(int keyCode)
	{
		ProcessKeys(keyCode);
    }
    
    /**
     * Called when a key is released.
     */
    protected  void keyReleased(int keyCode)
	{
    }

    /**
     * Called when a key is repeated (held down).
     */
    protected  void keyRepeated(int keyCode) {
    }
    
    /**
     * Called when the pointer is dragged.
     */
    protected  void pointerDragged(int x, int y) {
    }

    /**
     * Called when the pointer is pressed.
     */
    protected  void pointerPressed(int x, int y) {
    }

    /**
     * Called when the pointer is released.
     */
    protected  void pointerReleased(int x, int y) {
    }
    
    /**
     * Called when action should be handled
     */
    public void commandAction(Command command, Displayable displayable)
	{
    }

	private void forceRepaint()
	{
		forceRepaint = true;
		repaint();
	}
	
	private void ProcessKeys(int keyCode)
	{
		int gameAction = getGameAction(keyCode);
		switch (gameAction)
		{
			case UP:
				timeOffset = 0;
				forceRepaint();
				break;
			case DOWN:
				timeOffset = 0;
				forceRepaint();
				break;
			case LEFT:
				timeOffset -= 3600 * 24;
				forceRepaint();
				break;
			case RIGHT:
				timeOffset += 3600 * 24;
				forceRepaint();
				break;
			case FIRE:
				break;
			default:
				switch (keyCode)
				{
					case KEY_NUM2:
						timeOffset = 0;
						forceRepaint();
						break;
					case KEY_NUM8:
						timeOffset = 0;
						forceRepaint();
						break;
					case KEY_NUM4:
						timeOffset -= 3600 * 24;
						forceRepaint();
						break;
					case KEY_NUM6:
						timeOffset += 3600 * 24;
						forceRepaint();
						break;
					case KEY_NUM5:
						break;
					case KEY_NUM3:
						break;
					case KEY_NUM1:
						break;
					case KEY_NUM7:
						break;
					case KEY_STAR:
						break;
					case KEY_NUM0:
						break;
					case KEY_POUND:
						break;
					case KEY_NUM9:
						break;
					default:
						return;
				}
		}
	}
}